from django.core.files.storage import FileSystemStorage
from django.conf import settings


class DynamicFileSystemStorage(FileSystemStorage):
    def __init__(self, location=settings.DYNAMIC_FILE_STORAGE_LOCATION):
        super().__init__(location=location, base_url='/admin/download/')



from django.db import models
from django.utils.translation import gettext as _
from dynamic_file.models.base import DynamicFileBase
from dynamic_file.storage import DynamicFileSystemStorage


fs = DynamicFileSystemStorage()


class DynamicFile(DynamicFileBase):
    file = models.FileField(
        storage=fs,
        help_text=_('The uploaded file')
    )
    '''
    The concrete file for this model.
    '''



from django.contrib import admin

from .models import DynamicFile
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _


def preview(dynamic_file):
    try:
        mimetype = dynamic_file.mimetype
        if mimetype and 'image' in mimetype:
            src = dynamic_file.to_base64_src()
            return mark_safe(f'<img src="{src}" width="150" />')
        else:
            return _('No preview available')
    except Exception as e:
        return _(f'No preview available: {str(e)}')


@admin.register(DynamicFile)
class DynamicFileAdmin(admin.ModelAdmin):
    list_display = ['id', 'name']
    readonly_fields = ['preview']

    def preview(self, instance):
        return preview(instance)



from django.test import TestCase
from django.apps import apps
from django.conf import settings

from dynamic_file.models import DynamicFile

import helpers
import base64

UploadedByModel = apps.get_model(settings.DYNAMIC_FILE_UPLOADED_BY_MODEL)


class DynamicFileModelTestCase(TestCase):

    def test_create_default(self):
        instance = DynamicFile.objects.create()

        assert instance.description == ''
        assert instance.uploaded_by is None
        instance.file.name == ''

    def test_description(self):
        desc = 'A sample description for this file'
        instance = DynamicFile.objects.create(description=desc)

        assert instance.description == desc

    def test_name(self):
        instance = DynamicFile.objects.create(file=helpers.create_dummy_gif())
        assert instance.name == instance.file.name

    def test_storage_location(self):
        instance = DynamicFile.objects.create(file=helpers.create_dummy_gif())

        assert instance.file.name != ''
        assert instance.file.read() == helpers.create_dummy_gif().read()

    def test_read(self):
        instance = DynamicFile.objects.create(file=helpers.create_dummy_gif())
        assert instance.read == instance.file.read

    def test_mimetype(self):
        instance = DynamicFile.objects.create(file=helpers.create_dummy_gif('image.gif'))
        mimetype = instance.mimetype
        assert mimetype
        assert mimetype == 'image/gif'

    def test_to_base64(self):
        file = helpers.create_dummy_gif('image.gif')
        instance = DynamicFile.objects.create(file=file)

        assert base64.b64encode(helpers.SMALL_GIF) == instance.to_base64()

    def test_to_base64_utf8(self):
        file = helpers.create_dummy_gif('image.gif')
        instance = DynamicFile.objects.create(file=file)

        assert base64.b64encode(helpers.SMALL_GIF).decode('utf-8') == instance.to_base64_utf8()

    def test_to_base64_src(self):
        file = helpers.create_dummy_gif('image.gif')
        instance = DynamicFile.objects.create(file=file)

        b64 = base64.b64encode(helpers.SMALL_GIF).decode('utf-8')
        expected = f'data:;base64,{b64}'
        assert expected == instance.to_base64_src()

    def test_str(self):
        instance = DynamicFile.objects.create(file=None)
        assert str(instance.pk) in str(instance)

        instance = DynamicFile.objects.create(display_name='test')
        assert instance.display_name == str(instance)

        instance = DynamicFile.objects.create(file=helpers.create_dummy_gif('image.gif'))
        assert instance.name == str(instance)



# Generated by Django 4.1.3 on 2023-07-07 07:34

from django.db import migrations, models
import django.db.models.deletion


from django.conf import settings
DYNAMIC_FILE_UPLOADED_BY_MODEL = settings.DYNAMIC_FILE_UPLOADED_BY_MODEL


class Migration(migrations.Migration):

    dependencies = [
        ('dynamic_file', '0002_alter_dynamicfile_file'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dynamicfile',
            name='uploaded_by',
            field=models.ForeignKey(blank=True, help_text='The owner/uploader of this file', null=True,
                                    on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=DYNAMIC_FILE_UPLOADED_BY_MODEL),
        ),
        migrations.AddField(
            model_name='dynamicfile',
            name='display_name',
            field=models.CharField(blank=True, default='',
                                   help_text='An optional displayable name for this file.', max_length=128),
        ),
    ]



# Generated by Django 4.1.3 on 2022-11-16 19:28

from django.db import migrations, models
import django.db.models.deletion

from django.conf import settings

from dynamic_file import config

# Heavily inspired by https://github.com/dj-stripe/dj-stripe/blob/master/djstripe/migrations/0001_initial.py

DYNAMIC_FILE_UPLOADED_BY_MODEL = settings.DYNAMIC_FILE_UPLOADED_BY_MODEL
DYNAMIC_FILE_UPLOADED_BY_RELATED_NAME = '+'
DYNAMIC_FILE_UPLOADED_BY_MODEL_MIGRATION_DEPENDENCY = settings.DYNAMIC_FILE_UPLOADED_BY_MIGRATION_DEPENDENCY


DYNAMIC_FILE_UPLOADED_BY_MODEL_DEPENDENCY = migrations.swappable_dependency(
    DYNAMIC_FILE_UPLOADED_BY_MODEL
)

if DYNAMIC_FILE_UPLOADED_BY_MODEL != settings.AUTH_USER_MODEL:
    DYNAMIC_FILE_UPLOADED_BY_MODEL_DEPENDENCY = migrations.migration.SwappableTuple(
        (
            DYNAMIC_FILE_UPLOADED_BY_MODEL.split('.', 1)[0],
            DYNAMIC_FILE_UPLOADED_BY_MODEL_MIGRATION_DEPENDENCY,
        ),
        DYNAMIC_FILE_UPLOADED_BY_MODEL,
    )


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        DYNAMIC_FILE_UPLOADED_BY_MODEL_DEPENDENCY
    ]

    operations = [
        migrations.CreateModel(
            name='DynamicFile',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(blank=True, help_text='A description for this file')),
                ('file', models.FileField(help_text='The uploaded file', upload_to='')),
                ('uploaded_by', models.ForeignKey(help_text='The owner/uploader of this file', null=True,
                 on_delete=django.db.models.deletion.SET_NULL, related_name=DYNAMIC_FILE_UPLOADED_BY_RELATED_NAME, to=DYNAMIC_FILE_UPLOADED_BY_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.test import RequestFactory
from django.test import TestCase
from django.urls import reverse
from django.core.files.uploadedfile import SimpleUploadedFile

from rest_framework import serializers

from dynamic_file.models import DynamicFile
from dynamic_file.serializers import DynamicFileField

from test_app.models import TestModelOneToOne

import helpers


class ModelOneToOneSerializer(serializers.ModelSerializer):
    file = DynamicFileField(allow_null=True)

    def get_file_url(self, dynamic_file):
        return reverse('serve_default', kwargs={'pk': dynamic_file.id})

    class Meta:
        model = TestModelOneToOne
        fields = ['file']


class ModelOneToOneSerializerWithFallback(serializers.ModelSerializer):
    file = DynamicFileField()

    def get_file_fallback_url(self, instance):
        return instance.id  # just to mock something unique

    class Meta:
        model = TestModelOneToOne
        fields = ['file']


class ModelOneToOneSerializerWithArgument(serializers.ModelSerializer):
    file = DynamicFileField(view_name='serve_default', view_args={'pk': 'id'})

    class Meta:
        model = TestModelOneToOne
        fields = ['file']


class DynamicFileSerializerFieldTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.request = RequestFactory().get('', data={})

    def test_serialize_null(self):
        instance = TestModelOneToOne.objects.create()

        expected = {'file': None}
        actual = ModelOneToOneSerializer(instance=instance, context={'request': self.request}).data

        assert expected == actual

    def test_serialize_null_fallback(self):
        instance = TestModelOneToOne.objects.create()

        expected = {'file': instance.pk}
        actual = ModelOneToOneSerializerWithFallback(instance=instance, context={'request': self.request}).data

        assert expected == actual

    def test_serialize(self):
        file = DynamicFile.objects.create(file=helpers.create_dummy_gif())
        instance = TestModelOneToOne.objects.create(file=file)

        expected = {'file': f'http://testserver/serve/{file.id}'}
        actual = ModelOneToOneSerializer(instance=instance, context={'request': self.request}).data

        assert expected == actual

    def test_serialize_no_request(self):
        file = DynamicFile.objects.create(file=helpers.create_dummy_gif())
        instance = TestModelOneToOne.objects.create(file=file)

        expected = {'file': f'/serve/{file.id}'}
        actual = ModelOneToOneSerializer(instance=instance).data

        assert expected == actual

    def test_serialize_view_arg(self):
        file = DynamicFile.objects.create(file=helpers.create_dummy_gif())
        instance = TestModelOneToOne.objects.create(file=file)

        expected = {'file': f'/serve/{file.id}'}
        actual = ModelOneToOneSerializerWithArgument(instance=instance).data

        assert expected == actual

    def test_create(self):
        payload = {'file': helpers.create_dummy_gif()}
        serializer = ModelOneToOneSerializer(data=payload)
        assert serializer.is_valid()
        instance = serializer.save()

        assert instance.file
        assert instance.file.file.read() == helpers.create_dummy_gif().read()
        assert DynamicFile.objects.count() == 1

    def test_create_empty(self):
        payload = {'file': None}
        serializer = ModelOneToOneSerializer(data=payload)
        assert serializer.is_valid()
        instance = serializer.save()

        assert instance.file is None
        assert DynamicFile.objects.count() == 0

    def test_update(self):
        file = DynamicFile.objects.create(file=helpers.create_dummy_gif())
        instance = TestModelOneToOne.objects.create(file=file)

        payload = {'file': SimpleUploadedFile('anotherfile.txt', b'some_content')}
        serializer = ModelOneToOneSerializer(instance=instance, data=payload)
        assert serializer.is_valid()

        instance = serializer.save()
        assert instance.file.file.read() == b'some_content'
        assert DynamicFile.objects.count() == 1

    def test_delete(self):
        file = DynamicFile.objects.create(file=helpers.create_dummy_gif())
        instance = TestModelOneToOne.objects.create(file=file)

        payload = {'file': None}
        serializer = ModelOneToOneSerializer(instance=instance, data=payload)
        assert serializer.is_valid()

        instance = serializer.save()
        assert instance.file is None
        assert DynamicFile.objects.count() == 0



from django.db import models
from django.utils.translation import gettext as _
from django.conf import settings
import mimetypes

import base64


class DynamicFileBase(models.Model):
    '''
    Base model for handling dynamic files. Abstract, cannot be used directly.
    '''

    display_name = models.CharField(
        blank=True,
        default='',
        max_length=128,
        help_text=_('An optional displayable name for this file.')
    )
    '''
    A concise and optional description of the uploaded file.
    '''

    description = models.TextField(
        blank=True,
        help_text=_('A description for this file')
    )
    '''
    A concise and optional description of the uploaded file.
    '''

    uploaded_by = models.ForeignKey(
        settings.DYNAMIC_FILE_UPLOADED_BY_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name='+',
        help_text=_('The owner/uploader of this file')
    )
    '''
    Specifies the uploader of this file. This foreign key defaults to ``AUTH_USER_MODEL``
    but can be customized by changing ``DYNAMIC_FILE_UPLOADED_BY_MODEL``.
    The reverse accessor is unused.

    NOTE: Changing those settings is only supported _before_ running the first migration
    '''

    created_at = models.DateTimeField(
        auto_now_add=True,
        blank=True,
        help_text=_('The creation date of this file')
    )
    '''
    Read-only field, specifying the upload date of this file.
    '''

    updated_at = models.DateTimeField(
        auto_now=True,
        blank=True,
        help_text=_('The last update date of this file')
    )
    '''
    Read-only field, specifying the update date of this file.
    '''

    @property
    def name(self):
        '''
        Returns the filename of this instance. If ``file`` is ``None``, an empty string is returned.
        '''
        if self.file:
            return self.file.name
        return ''

    @property
    def read(self):
        '''
        Wrapper for the read method of the ``FileField`` instance
        '''
        return self.file.read

    @property
    def mimetype(self):
        '''
        Guesses the ``mimetype`` from the file extension. Returns null if unknown
        '''
        return mimetypes.MimeTypes().guess_type(self.name)[0]

    def to_base64(self):
        '''
        Converts the file to base64. Undefined if the file is ``None``
        '''
        return base64.b64encode(self.read())

    def to_base64_utf8(self):
        '''
        Converts the file to base64 and returns it in ``utf-8`` encoding.
        '''
        return self.to_base64().decode('utf-8')

    def to_base64_src(self):
        '''
        Converts the file to base64 and returns it in ``utf-8`` encoding. In addition,
        the content is wrapped to be used within html ``src=`` attributes
        '''
        src = self.to_base64_utf8()
        return f'data:;base64,{src}'

    def __str__(self):
        '''
        String representation of this model.
        If a display name is set, this will be used as representation.
        If no display name is set, the filename is used.
        If no filename is set, the pk and a `_nofile` suffix is used.
        '''
        if len(self.display_name) > 0:
            return self.display_name
        elif len(self.name) > 0:
            return self.name
        else:
            return f'{self.pk}_nofile'

    class Meta:
        abstract = True



from django.http import FileResponse
from django.http import HttpResponse
from django.test import TestCase
from django.test import override_settings
from django.conf import settings

from rest_framework.test import APIRequestFactory
from rest_framework import status

from dynamic_file.models import DynamicFile
from dynamic_file.views import ServeDynamicFile

from test_app.views import ServeTestFile

import os

import helpers


class ServeDynamicFileTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.view = ServeDynamicFile.as_view()
        cls.instance_1 = DynamicFile.objects.create(file=helpers.create_dummy_gif())

    @override_settings(DEBUG=True)
    def test_serve_content_debug(self):
        factory = APIRequestFactory()
        request = factory.get('')

        response = self.view(request, pk=self.instance_1.pk)

        assert response.status_code is status.HTTP_200_OK
        assert isinstance(response, FileResponse)
        assert response.filename == self.instance_1.file.name

    def test_serve_content_production(self):
        factory = APIRequestFactory()
        request = factory.get('')

        response = self.view(request, pk=self.instance_1.pk)
        expected_path = os.path.join(settings.DYNAMIC_FILE_STORAGE_LOCATION, self.instance_1.file.name)

        assert response.status_code is status.HTTP_200_OK
        assert isinstance(response, HttpResponse)
        assert 'X-Accel-Redirect' in response.headers.keys()
        assert 'Content-Type' in response.headers.keys()
        assert 'Content-Disposition' in response.headers.keys()
        assert 'Content-Encoding' in response.headers.keys()
        assert response.headers['X-Accel-Redirect'] == expected_path

        with open(response.headers['X-Accel-Redirect'], 'rb') as file:
            assert file.read() == self.instance_1.file.read()

    def test_not_found(self):
        factory = APIRequestFactory()
        request = factory.get('')

        response = self.view(request, pk=-1)

        assert response.status_code is status.HTTP_404_NOT_FOUND


class ServeCustomDynamicFileTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.view = ServeTestFile.as_view()
        cls.instance_1 = DynamicFile.objects.create(file=helpers.create_dummy_gif())

    @override_settings(DEBUG=True)
    def test_serve_content(self):
        factory = APIRequestFactory()
        request = factory.get('')

        response = self.view(request, key=self.instance_1.pk)

        assert response.status_code is status.HTTP_200_OK
        assert isinstance(response, FileResponse)
        assert response.filename == self.instance_1.file.name



from django.urls import reverse

from rest_framework import serializers

from dynamic_file.models import DynamicFile


class _DynamicFileFieldMixin():
    _Model = DynamicFile()
    _view_name = None
    _view_args = {}

    def get_attribute(self, instance):
        value = super().get_attribute(instance)
        return value if value else '__invalid__'

    def to_representation(self, dynamic_file):
        if dynamic_file == '__invalid__':
            method_name = f'get_{self.field_name}_fallback_url'
            method = getattr(self.parent, method_name, None)
            return method(self.parent.instance) if method else None

        if self._view_name:
            args = {}
            for k, v in self._view_args.items():
                args[k] = getattr(dynamic_file, v)
            rep = reverse(self._view_name, kwargs=args)
        else:
            method_name = f'get_{self.field_name}_url'
            method = getattr(self.parent, method_name)
            rep = method(dynamic_file)

        request = self.context.get('request', None)
        return request.build_absolute_uri(rep) if request else rep

    def run_validation(self, data):
        value = self.to_internal_value(data)
        self.run_validators(value)
        return value

    def to_internal_value(self, data):
        instance = self.parent.instance

        if data is None:
            asset = getattr(instance, self.field_name, None) if instance else None
            if asset:
                asset.delete()
            return None
        elif instance is None or getattr(instance, self.field_name, None) is None:
            asset = self._Model.objects.create(file=data)
            return asset
        else:
            asset = getattr(instance, self.field_name)
            asset.file = data
            asset.save()
            return asset


class DynamicFileField(_DynamicFileFieldMixin, serializers.FileField):
    _Model = DynamicFile

    def __init__(
        self,
        view_name=None,
        view_args={},
        *args,
        **kwargs
    ):
        kwargs['use_url'] = False
        self._view_name = view_name
        self._view_args = view_args
        super().__init__(*args, **kwargs)




from django.test import TestCase

from dynamic_file.models import DynamicFile
from dynamic_file.admin import preview
import helpers


class AdminTestCase(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.instance_1 = DynamicFile.objects.create(file=helpers.create_dummy_gif())
        cls.instance_2 = DynamicFile.objects.create(file=helpers.create_dummy_gif('not_a_gif.exe'))

    def test_preview(self):
        response = preview(self.instance_1)
        assert '<img' in response

    def test_preview_no_image(self):
        response = preview(self.instance_2)
        assert '<img' not in response



from django.apps import AppConfig

from dynamic_file import config

config.settings


class DynamicFileConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'dynamic_file'



# Generated by Django 4.1.3 on 2023-05-22 16:21

from django.db import migrations, models
import dynamic_file.storage


class Migration(migrations.Migration):

    dependencies = [
        ('dynamic_file', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='dynamicfile',
            name='file',
            field=models.FileField(help_text='The uploaded file', storage=dynamic_file.storage.DynamicFileSystemStorage(), upload_to=''),
        ),
    ]



# Generated by Django 4.1.3 on 2023-07-07 11:35

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('dynamic_file', '0003_alter_dynamicfile_uploaded_by'),
    ]

    operations = [
        migrations.AddField(
            model_name='dynamicfile',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, help_text='The creation date of this file'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='dynamicfile',
            name='updated_at',
            field=models.DateTimeField(auto_now=True, help_text='The last update date of this file'),
        ),
    ]



from django.conf import settings
from django.http.response import FileResponse
from django.http.response import HttpResponse

from rest_framework.views import APIView
from rest_framework import status
from rest_framework.exceptions import NotFound

from dynamic_file.models import DynamicFile

from rest_framework.settings import api_settings
from rest_framework.permissions import IsAuthenticated
from rest_framework.permissions import IsAdminUser

import mimetypes
import os


class _DynamicContentMixin():
    _Model = DynamicFile
    permission_classes = api_settings.DEFAULT_PERMISSION_CLASSES

    lookup_field = 'pk'
    lookup_url_kwarg = 'pk'

    def get_object(self):
        filter_kwargs = {self.lookup_field: self.kwargs[self.lookup_url_kwarg]}
        instance = self._Model.objects.filter(**filter_kwargs).first()
        self.check_object_permissions(self.request, instance)
        return instance


class ServeDynamicFile(_DynamicContentMixin, APIView):
    def get(self, request, *args, **kwargs):
        instance = self.get_object()
        filename = None
        if instance:
            filename = instance.file.name
        else:
            raise NotFound()

        content_type, encoding = mimetypes.guess_type(filename)
        path = os.path.join(settings.DYNAMIC_FILE_STORAGE_LOCATION, filename)

        try:
            if settings.DEBUG:
                fp = open(path, 'rb')
                response = FileResponse(fp)
                response.filename = filename

            else:  # for now, nginx will suffice
                response = HttpResponse(content_type=content_type)
                response['Content-Disposition'] = 'inline; filename={0}'.format(filename)
                response['Content-Encoding'] = encoding

                location = os.path.normpath(path)
                response['X-Accel-Redirect'] = location

            return response
        except Exception as e:
            print(e)
            return HttpResponse('File not found', status=status.HTTP_404_NOT_FOUND)


class ServeDynamicFileAdmin(ServeDynamicFile):
    permission_classes = [IsAuthenticated, IsAdminUser]
    lookup_field = 'file'
    lookup_url_kwarg = 'name'



from django.conf import settings

if not hasattr(settings, 'DYNAMIC_FILE_UPLOADED_BY_MODEL'):
    setattr(settings, 'DYNAMIC_FILE_UPLOADED_BY_MODEL', settings.AUTH_USER_MODEL)

if not hasattr(settings, 'DYNAMIC_FILE_STORAGE_LOCATION'):
    setattr(settings, 'DYNAMIC_FILE_STORAGE_LOCATION', 'files')

if not hasattr(settings, 'DYNAMIC_FILE_UPLOADED_BY_MIGRATION_DEPENDENCY'):
    setattr(settings, 'DYNAMIC_FILE_UPLOADED_BY_MIGRATION_DEPENDENCY', '__first__')
